from pandas.core.common import flatten
from tqdm import trange
import torch
import numpy as np


theo_clonal_list = {
    "1:0" : 1,
    "1:1" : 1,
    "2:0" : 2,
    "2:1" : 2,
    "2:2" : 2
}

theo_allele_list = {
    "1:0" : 1,
    "1:1" : 2,
    "2:0" : 2,
    "2:1" : 3,
    "2:2" : 4
}

theo_clonal_means_list = {
    "1:0" : torch.tensor(0.95),
    "1:1" : torch.tensor(0.5),
    "2:0" : torch.tensor([0.5,0.95]),
    "2:1" : torch.tensor([0.333,0.666]),
    "2:2" : torch.tensor([0.25,0.5])
}

def flatten_list(l):
    return list(flatten(l))


def log_sum_exp(args):
    c = torch.amax(args, dim=0)
    return c + torch.log(torch.sum(torch.exp(args - c), axis=0))


def get_theo_clones(data):
    karyos = list(data.keys())
    theoretical_num_clones = [theo_clonal_list[kr] for kr in karyos]
    return theoretical_num_clones

def get_clones_counts(theoretical_num_clones):
    uniques = list(set(theoretical_num_clones))
    uniques = {k : 0 for k in uniques}
    res = [None] * len(theoretical_num_clones)
    for i in range(len(theoretical_num_clones)):
        res[i] = uniques[theoretical_num_clones[i]]
        uniques[theoretical_num_clones[i]] += 1
    return(res)


def compute_entropy(params, tail):

    res = 0
    posts = params["cluster_probs"]
    for k in posts:
        posts_k = posts[k]
        if tail:
            posts_k = posts_k[1:]
        log_post_k = torch.log(posts_k + 0.000001)
        post_k_entr = posts_k * log_post_k
        post_k_entr = torch.sum(post_k_entr, axis = 0)
        post_k_entr = -1 * torch.sum(post_k_entr)
        res += post_k_entr
    return res



def format_parameters_for_export(data, params, tail):
    res = {k : 0 for k in data.keys()}
    theoretical_num_clones = get_theo_clones(data)
    clones_count = get_clones_counts(theoretical_num_clones)
    for i, k in enumerate(res):
        res[k] = format_parameters_for_export_aux(data, params,k, i, theoretical_num_clones, clones_count, tail)
    return res


def format_parameters_for_export_aux(data, params,k, i, theo_clones, counts_clone, tail):

    j = counts_clone[i]
    if theo_clones[i] == 2:
        beta_concentration1 = params['a_2'][:,j] * params['b_2'][:,j]
        beta_concentration2 = (1 - params['a_2'][:,j]) * params['b_2'][:, j]
    else:
        beta_concentration1 = params['a_1'][:, j] * params['b_1'][:, j]
        beta_concentration2 = (1 - params['a_1'][:, j]) * params['b_1'][:, j]

    mixture_weights = params['param_weights_{}'.format(theo_clones[i])][j, :].detach().numpy()
    if tail == 1:
        tail_weights = params['param_tail_weights'][i, :].detach().numpy()
        mixture_weights = mixture_weights * tail_weights[1]
        mixture_weights = np.insert(mixture_weights, 0, tail_weights[0])


    ca, order_vec = rename_clusters(np.argmax(params["cluster_probs"][k].detach().numpy(), axis = 0), tail, theo_clones[i])

    res = {"cluster_probs" : params["cluster_probs"][k].detach().numpy(),
           "cluster_assignments" : ca,
           "cluster_types" : order_vec,
           "mixture_probs" : mixture_weights,
          "beta_concentration1" : beta_concentration1.detach().numpy(),
          "beta_concentration2" : beta_concentration2.detach().numpy()
        }
    if tail == 1:
        res["tail_shape"] = params['tail_mean'].detach().numpy()
        res["tail_scale"] = np.min(data[k].detach().numpy())
        res["tail_noise"] =  1/params['alpha_noise'][i].detach().numpy()

    return res


def rename_clusters(x,tail, theo_c):
    res = np.empty(len(x), dtype='object')
    uniq_clusts = np.unique(x)
    base_idx = 0
    clonal_num = 1
    subclonal_num = 1
    order_vec = np.array([],dtype='object' )
    ###  IDENTIFY TAIL MUTATIONS ###
    if tail == 1:
        res[x == 0] = "Tail"
        np.append(order_vec,"Tail")
        base_idx += 1

    ### IDENTIFY CLONAL MUTATIONS ###
    for i in range(base_idx, base_idx + theo_c):
        res[x == i] = "C" + str(clonal_num)
        np.append(order_vec,"C" + str(clonal_num))
        clonal_num += 1

    ### IDENTIFY SUBCLONAL MUTATIONS ###
    for i in range(base_idx + theo_c,len(uniq_clusts)):
        res[x == i] = "S" + str(subclonal_num)
        np.append(order_vec, "S" + str(subclonal_num))
        subclonal_num += 1

    return res,order_vec




import pyro as pyro
import numpy as np
from pyro.infer import MCMC, NUTS, Predictive, HMC, config_enumerate
from pyro.infer.autoguide import AutoDelta, init_to_sample
import pyro.distributions as dist
import pyro.poutine as poutine
import torch
from torch.distributions import constraints
import mobster.utils_mobster as mut


@config_enumerate
def guide(data, K=1, tail=1, purity=0.96, clonal_beta_var=1., number_of_trials_clonal_mean=100.,
          number_of_trials_clonal=900., number_of_trials_k=300., prior_lims_clonal=[1., 10000.],alpha_precision_concentration = 5, alpha_precision_rate=0.1,
          prior_lims_k=[1., 10000.]):
    """
    Parameters
    ----------
    param1 :
    The first parameter.
    param2 :
    The second parameter.

    """
    karyos = list(data.keys())

    # Here we calculate the theoretical number of clonal clusters
    theoretical_num_clones = [mut.theo_clonal_list[kr] for kr in karyos]

    # Calculate the theoretical clonal means, wihch can be analytically defined for simple karyotypes, and then multiply them by the ploidy
    theoretical_clonal_means = [mut.theo_clonal_means_list[kr] * purity for kr in karyos]

    counts_clones = dict()
    for i in theoretical_num_clones:
        counts_clones[i] = counts_clones.get(i, 0) + 1

    index_2 = [i for i, j in enumerate(theoretical_num_clones) if j == 2]
    index_1 = [i for i, j in enumerate(theoretical_num_clones) if j == 1]

    a_prior = pyro.param("tail_mean", torch.tensor(1.5), constraint=constraints.interval(0.5, 5))
    alpha_precision_par = pyro.param("alpha_noise",
                                 dist.Gamma(concentration=alpha_precision_concentration, rate=alpha_precision_rate).mean * torch.ones([len(karyos)]),
                                 constraint=constraints.positive)





    alpha_prior = pyro.sample('u', dist.Delta(a_prior))

    idx1 = 0
    idx2 = 0

    if tail == 1:
        weights_tail = pyro.param("param_tail_weights", 1 / torch.ones([len(karyos), 2]), constraint=constraints.simplex)

    if 2 in theoretical_num_clones:
        weights_param_2 = pyro.param("param_weights_2", (1 / (K + 2)) * torch.ones([len(index_2), K + 2]),
                                     constraint=constraints.simplex)
    if 1 in theoretical_num_clones:
        weights_param_1 = pyro.param("param_weights_1", (1 / (K + 1)) * torch.ones([len(index_1), K + 1]),
                                     constraint=constraints.simplex)

    if K != 0:
        if 2 in counts_clones.keys():

            k2 = torch.cat([theoretical_clonal_means[kr] for kr in range(len(karyos)) if theoretical_num_clones[kr] == 2]).reshape([counts_clones[2],2])


            k_means_2 = ((torch.amin(k2, dim = 1) - 0.05) / (K +1) ).reshape([1, counts_clones[2]]) * torch.arange(1, K + 1).reshape([K,1]) + 0.12
        else:
            k_means_2 = 0


        if 1 in counts_clones.keys():
            k1 = torch.tensor([theoretical_clonal_means[kr] for kr in range(len(karyos)) if theoretical_num_clones[kr] == 1]).reshape([counts_clones[1],1])

            k_means_1 = ((torch.amin(k1, dim = 1) - 0.05) / (K + 1)).reshape([1,counts_clones[1]]) * torch.arange(1, K + 1).reshape([K, 1])  + 0.12
        else:
            k_means_1 = 0
    else:
        k_means_1 = k_means_2 = 0



    for kr in pyro.plate("kr", len(karyos)):



        if theoretical_num_clones[kr] == 2:

            pyro.sample('weights_{}'.format(kr), dist.Delta(weights_param_2[idx2]).to_event(1))

            # Mean parameter when the number of clonal picks is 2
            a_2_theo = torch.cat([theoretical_clonal_means[i] for i in index_2]).reshape([2, counts_clones[2]])

            # Number of trials parameter when the number of clonal picks is 2
            b_2_theo = torch.ones([2, len(index_2)]) * number_of_trials_clonal_mean

            # If we have subclones
            a21 = a22 = 0
            if K != 0:
                # mean of the subclones, located more or less between the smallest clonal cluster and the tail
                a_2_k = dist.Uniform(k_means_2 - 0.05, k_means_2 + 0.05).sample().reshape([K, counts_clones[2]])
                # Number of trials  for the subclones
                b_2_k = torch.ones([K, len(index_2)]) * number_of_trials_k

                a21 = pyro.param('a_2',
                                 torch.cat((a_2_theo, a_2_k)).reshape([2 + K, len(index_2)]),
                                 constraint=constraints.unit_interval)
                a22 = pyro.param('b_2',
                                 torch.cat((b_2_theo, b_2_k)).reshape([2 + K, len(index_2)]),
                                 constraint=constraints.positive)
            else:
                a21 = pyro.param('a_2',
                                 a_2_theo.reshape([2 + K, len(index_2)]),
                                 constraint=constraints.unit_interval)
                a22 = pyro.param('b_2',
                                 b_2_theo.reshape([2 + K, len(index_2)]),
                                 constraint=constraints.positive)



            with pyro.plate("clones_{}".format(kr), 2 + K):
                pyro.sample('beta_clone_mean_{}'.format(kr), dist.Delta(a21[:, idx2]))
                pyro.sample('beta_clone_n_samples_{}'.format(kr), dist.Delta(a22[:, idx2]))
            idx2 += 1

        else:
            pyro.sample('weights_{}'.format(kr), dist.Delta(weights_param_1[idx1]).to_event(1))

            # Mean parameter when the number of clonal picks is 1
            a_1_theo = torch.tensor([theoretical_clonal_means[i] for i in index_1]).reshape([1, counts_clones[1]])

            # Number of trials parameter when the number of clonal picks is 1
            b_1_theo = torch.ones([1, len(index_1)]) * number_of_trials_clonal_mean

            if K != 0:

                a_1_k = dist.Uniform(k_means_1 - 0.05, k_means_1 + 0.05).sample().reshape([K, counts_clones[1]])
                b_1_k = torch.ones([K, len(index_1)]) * number_of_trials_k

                a11 = pyro.param('a_1',
                                 torch.cat((a_1_theo, a_1_k)).reshape([1 + K, len(index_1)]),
                                 constraint=constraints.unit_interval)
                a12 = pyro.param('b_1',
                                 torch.cat((b_1_theo, b_1_k)).reshape([1 + K, len(index_1)]),
                                 constraint=constraints.positive)
            else:


                a11 = pyro.param('a_1',
                                 a_1_theo.reshape([1 + K, len(index_1)]),
                                 constraint=constraints.unit_interval)
                a12 = pyro.param('b_1',
                                 b_1_theo.reshape([1 + K, len(index_1)]),
                                 constraint=constraints.positive)

            with pyro.plate("clones_{}".format(kr), 1 + K):
                pyro.sample('beta_clone_mean_{}'.format(kr), dist.Delta(a11[:, idx1]))
                pyro.sample('beta_clone_n_samples_{}'.format(kr), dist.Delta(a12[:, idx1]))
            idx1 += 1

        if tail == 1:
            # K = K + tail
            pyro.sample('weights_tail_{}'.format(kr), dist.Delta(weights_tail[kr]).to_event(1))
            alpha_precision = pyro.sample('alpha_precision_{}'.format(kr), dist.Delta(alpha_precision_par[kr]))

            with pyro.plate('data_{}'.format(kr), len(data[karyos[kr]])):
                pyro.sample("alpha_noise_{}".format(kr),
                            dist.LogNormal(torch.log(alpha_prior) * mut.theo_allele_list[karyos[kr]], 1 / alpha_precision))




from pyro.distributions import Rejector, Pareto
import torch



class BoundedPareto(Rejector):
    def __init__(self, scale, alpha, upper_limit, validate_args=None):
        propose = Pareto(scale, alpha, validate_args=validate_args)

        def log_prob_accept(x):
            return (x < upper_limit).type_as(x).log()

        #log_scale = torch.Tensor(alpha) * torch.log(torch.Tensor([scale / upper_limit]))
        log_scale = 0
        super(BoundedPareto, self).__init__(propose, log_prob_accept, log_scale)